// U6Encode2.2.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"


// Ultima 6 compression utility
// Last updated on 18-February-2005

#include <iostream>
#include <stdio.h>

// -----------------------
// constants and variables
// -----------------------

const int block_size = 64;
int source_buffer_pos = 0;
int bits_written = 0;   // number of bits written to the dest buffer
unsigned char *source_buffer;
unsigned char *destination_buffer;
int source_buffer_size;
int destination_buffer_size;


// ---------------------
// function declarations
// ---------------------

bool end_of_input();
unsigned char get_input_byte();
void write9(int nbyte);
long get_file_size(FILE* file);
void write4(FILE *f, long outval);

void get_compressed_size (char* file_name);
int compress_file(char* input_file_name, char* output_file_name);

// -------------------
// auxiliary functions
// -------------------

bool end_of_input() {
   if (source_buffer_pos < source_buffer_size) return (false);
   else return (true);
}

unsigned char get_input_byte() {
   unsigned char input_byte = source_buffer[source_buffer_pos];
   source_buffer_pos++;
   return (input_byte);
}

void write9(int nbyte) {
   int ignore_bytes = bits_written / 8;
   int ignore_bits = bits_written % 8;
   int shifted_byte = nbyte << ignore_bits;
   unsigned char lo_byte = (unsigned char) (shifted_byte & 0xFF);
   unsigned char hi_byte = (unsigned char) ((shifted_byte >> 8) & 0xFF);
   unsigned char bit_mask_lo = ~((0xFF << ignore_bits) & 0xFF);
   unsigned char bit_mask_hi = (0xFF << (1+ignore_bits)) & 0xFF;

   // first byte
   destination_buffer[ignore_bytes+0]
      = (destination_buffer[ignore_bytes+0] & bit_mask_lo) | lo_byte;

   // second byte
   destination_buffer[ignore_bytes+1]
      = (destination_buffer[ignore_bytes+1] & bit_mask_hi) | hi_byte;

   bits_written += 9;
}

long get_file_size(FILE *file)
{
   long file_length;
   fseek(file, 0, SEEK_END);
   file_length = ftell(file);
   fseek(file, 0, SEEK_SET);
   return(file_length);
}

void write4(FILE *file, long outval)
{
   unsigned char out_array[4];
   out_array[0] = (unsigned char) (outval & 0xFF);
   out_array[1] = (unsigned char) ((outval >> 8) & 0xFF);
   out_array[2] = (unsigned char) ((outval >> 16) & 0xFF);
   out_array[3] = (unsigned char) ((outval >> 24) & 0xFF);
   fwrite(out_array, 1, sizeof(out_array), file);
}


// -------------------------
// the two central functions
// -------------------------

void get_compressed_size (char* file_name)
{
}

int compress_file(char* source_file_name, char* destination_file_name) {
   FILE* source_file;
   FILE* destination_file;

   if (strcmp (source_file_name, destination_file_name) == 0) {
      std::cout << "Source and destination cannot be identical.\n";
      return(EXIT_FAILURE);
   } else {
      if (!(source_file=fopen(source_file_name,"rb"))) {
         std::cout << "Could not open " << source_file_name;
         std::cout << ".\n";
         return(EXIT_FAILURE);
      } else {
         // allocate input buffer
         source_buffer_size = get_file_size(source_file);
         source_buffer =
            new unsigned char [source_buffer_size];

         // allocate output buffer
         int number_of_blocks =
            (source_buffer_size + (block_size-1)) / block_size;
         int destination_buffer_size =
            ((source_buffer_size * 9 + number_of_blocks * 9 + 1*9) + 7) / 8;
         destination_buffer =
    //        new unsigned char [source_buffer_size];
	        new unsigned char [destination_buffer_size];

         // load source file into input buffer
         fseek(source_file, 0, SEEK_SET);
         fread(source_buffer, 1, source_buffer_size, source_file);
		 fclose(source_file);

         // perform "compression" algorithm
         if (source_buffer_size > 0) {
            int bytes_processed;
            while (!end_of_input()) {
               write9(0x100);
               bytes_processed = 0;

               while ( (bytes_processed < block_size) && (!end_of_input()) ) {
                  write9(get_input_byte());
                  bytes_processed++;
               }
            }
            write9 (0x101);
         }

		 // save output buffer into destination file
         if (!(destination_file=fopen(destination_file_name,"wb"))) {
            std::cout << "Could not open " << destination_file_name;
            std::cout << ".\n";
            return(EXIT_FAILURE);
		 }
		 // write size header to output buffer
         write4 (destination_file, source_buffer_size);
         fwrite(destination_buffer, 1, destination_buffer_size,
                destination_file);
         fclose(destination_file);
      }
   }
}



// -------------
// main function
// -------------

int main (int argc, char *argv[]) {
   switch (argc) {
      case 2:
         get_compressed_size(argv[1]);
         return(EXIT_SUCCESS);
         break;
      case 3:
         compress_file(argv[1],argv[2]);
         return(EXIT_SUCCESS);
         break;
      default:
         std::cout << "Usage:\n";
         std::cout << "0 parameters - this message.\n";
         std::cout << "1 parameter  - display compressed file size.\n";
         std::cout << "2 parameters - compress arg1 into arg2.\n";
         return(EXIT_SUCCESS);
         break;
   }

   return 0;
}


